require 'fileutils'
require 'shellwords'
require 'socket'

RUBY_SCRIPT_HEADER=<<-EOT
#!/bin/bash
exec 2>&1
[ -e /etc/profile.d/rvm.sh ] && . /etc/profile.d/rvm.sh
EOT

RUBY_SCRIPT_HEADER_NOEXEC=<<-EOT
#!/bin/bash
[ -e /etc/profile.d/rvm.sh ] && . /etc/profile.d/rvm.sh
EOT

class PublishToWeb

  def disabled?
    File.exists?(config_file_path("publish_to_web_disabled")) || !File.exists?(config_file_path("license"))
  end

  def disable
    FileUtils.touch(config_file_path("publish_to_web_disabled"))
  end

  def enable
    FileUtils.rm(config_file_path("publish_to_web_disabled")) if disabled?
  end

  def config_file_path(name)
    dir = ENV["PROTONET_CONFIG"] || "/tmp/protonet-config/"
    FileUtils.mkdir_p(dir)
    "#{dir}/#{name}"
  end

  def call(cmd)
    require 'json'
    result = `#{cmd}`
    puts [cmd, result, "*"*50].join("\n") if ENV["DEBUG"]

    result = JSON.parse(result)
    raise ArgumentError.new(result['errors']) if result['errors']
    result
  end

  def get_harware_id
    IO.read(config_file_path("hardware_id")) rescue begin
      case RUBY_PLATFORM
      when /darwin/
        `ifconfig en0 | grep ether | awk '{ print $NF}'`
      when /linux/
        `ifconfig eth0 | grep -i hwaddr | awk '{ print $NF }'`
      else
        raise "unsupported platform"
      end.strip
    end
  end

  def set_public_key
    call "curl -s -X POST --form license_key=#{license} --form public_key='#{pubkey}' https://directory.protonet.info/set_public_key"
  end

  def pubkey
    IO.read("#{config_file_path("publish_to_web_key")}.pub").strip rescue nil
  end

  def regenerate_key
    FileUtils.rm_f [config_file_path("publish_to_web_key"), config_file_path("publish_to_web_key")+".pub"]
    puts "ssh-keygen -t dsa -f #{config_file_path("publish_to_web_key")} -N '' -C ''"
    `ssh-keygen -t dsa -f #{config_file_path("publish_to_web_key")} -N '' -C ''`
    set_public_key
  end

  def local_port
    IO.read(config_file_path("application_tcp_port")).strip rescue "80"
  end

  def local_port=(port)
    IO.write(config_file_path("application_tcp_port"), port)
  end

  def find_local_port
    server = TCPServer.new('127.0.0.1', 0)
    port = server.addr[1].to_s
    server.close
    port
  end

  def start_tunnel(ssh_identity, port_on_admin, port = nil)
    local_port = port if port

    puts "settings: #{info}"
    puts "starting tunnel to #{local_port}"
    port = find_local_port
    File.open(config_file_path("local_check_port"), "w") { |f| f.write port }

    Kernel::exec "/usr/bin/ssh localtunnel@directory.protonet.info \
      -o ServerAliveInterval=5                            \
      -o StrictHostKeyChecking=no                         \
      -o ExitOnForwardFailure=yes                         \
      -o UserKnownHostsFile=/dev/null                     \
      -p 22666                                            \
      -R #{port_on_admin}:127.0.0.1:#{local_port}         \
      -L #{port}:127.0.0.1:8765                           \
      -i #{ssh_identity}                                  \
      -N"
  end

  def create_license
    info = call "curl -s https://directory.protonet.info/create_license?hardware_id=#{get_harware_id}"
    info['license_key']
  end

  def copy_old_license
    IO.write(config_file_path("license"), IO.read("/home/protonet/.protonet-license").strip)
  rescue Errno::ENOENT => e
    false
  end

  def license
    copy_old_license
    IO.read(config_file_path("license")).strip
  rescue Errno::ENOENT => e
    IO.write(config_file_path("license"), create_license)
    retry
  end

  def version
    "#{File.basename(File.readlink("/protonet/firmware/current"))}/#{IO.read("/protonet/firmware/current/BUILD").strip}/#{File.readlink("/home/protonet/dashboard/current").split("-").last}"
  rescue
    "unkown"
  end

  def set_version
    call "curl -s -X POST --form license_key=#{license} --form version=#{version.shellescape} https://directory.protonet.info/set_version"
  end

  def info
    info = call "curl -s https://directory.protonet.info/info?license_key=#{license}"
    IO.write(nodename_file, info["node_name"])
    [info["port"], info["node_name"], info['pubkey_sha1']]
  end

  def nodename_file
    File.join(nodenames_path, 'publish_to_web')
  end

  def nodenames_path
    path = File.join(config_file_path(nil),'nodenames')
    FileUtils.mkdir_p path
    path
  end

  def key_ok?(pubkey_sha1)
    require 'digest/sha1'
    pubkey && (Digest::SHA1.hexdigest(pubkey) == pubkey_sha1)
  end

  def start(application_port = nil)
    while disabled?
      sleep 10
    end

    port, node_name, pubkey_sha1 = info

    regenerate_key unless key_ok?(pubkey_sha1)

    set_version

    start_tunnel(config_file_path("publish_to_web_key"), port, application_port)
  end

  def set_node_name(name)
    call "curl -s -X POST --form license_key=#{license} --form node_name=#{name.shellescape} https://directory.protonet.info/set_node_name"
    # update local .protonet-nodename
    info
  end

  def node_name
    port, node_name, pubkey_sha1 = info
    node_name
  end

  def check
    if disabled?
      puts "DISABLED"
      return true
    end

    port = IO.read(config_file_path("local_check_port")).to_i
    TCPSocket.new "127.0.0.1", port
    puts "OK"
    true
  rescue Errno::ECONNREFUSED
    puts "FAIL"
    false
  end

  def install(runit_service_directory, log_path, service_name = "publish_to_web")
    raise "service exists" if File.directory?("#{runit_service_directory}/#{service_name}")

    FileUtils.mkdir_p("#{runit_service_directory}/#{service_name}")

    File.open("#{runit_service_directory}/#{service_name}/run", "w", 0777) do |f|
      f.write "#{RUBY_SCRIPT_HEADER}\nexec publish_to_web start"
    end

    File.open("#{runit_service_directory}/#{service_name}/check", "w", 0777) do |f|
      f.write "#{RUBY_SCRIPT_HEADER_NOEXEC}\nexec publish_to_web check"
    end

    FileUtils.mkdir_p("#{log_path}/#{service_name}")

    FileUtils.mkdir_p("#{runit_service_directory}/#{service_name}/log")
    File.open("#{runit_service_directory}/#{service_name}/log/run", "w", 0777) do |f|
      f.write "#!/bin/bash\nexec svlogd -tt #{log_path}/#{service_name}"
    end
  end
end
